import numpy as np
import sasktran as sk
from tqdm import tqdm
from typing import List, Union


class AirMassFactorBase(object):
    """
    Base class for Air Mass Factor (AMF) classes using the SASKTRAN radiative transfer model.

    Each AMF class carries a weighting function (WF) object that holds most of the calculation parameters and does the
    radiative transfer calculations.
    """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, amf_species: str,
                 box_boundaries: np.array=np.linspace(0, 1e5, 101), box_midpoints: Union[None, np.array]=None,
                 wf_widths: Union[None, np.array]=None, total_sum=False, total_remove=False, **options):
        """
        Parameters
        ----------
        geometry : sasktran.Geometry
            Sasktran geometry object.
        atmosphere : sasktran.Atmosphere
            Sasktran atmosphere object.
        amf_species : str
            The name of the species the AMFs will be calculated for.
        box_boundaries : np.array
            Array of altitudes in metres defining the boundaries of the layers the box-AMFs will be calculated for.
            Default 1 km boxes from 0 to 100 km.
        box_midpoints : None, np.array
            Array of altitudes in metres defining the point within each layer the box-AMF will be calculated at.
            Defaults to the midpoint of the layer when box_boundaries is None.
        wf_widths : None, np.array
            The width in metres used to calculate the weighting functions used to calculate the box-AMFs. The width must
            be smaller than the distance between the lowest box_midpoint and the surface_height and smaller than the highest
            midpoint and the top-of-atmosphere in order to avoid unwanted interpolation effects, but it must be large
            enough to avoid numerical instability. Defaults to half the width of each layer when wf_widths is None.
        total_sum : False, bool
            True to calculate the total AMF by taking an average of the box AMFs weighted by the box vertical column.
            Requires no additional radiance calculations.
        total_remove : False, bool
            True to estimate the total AMF by calculating the radiance after removing the absorber. Only accurate
            for optically thin species. Requires two additional radiance calculations.
        **options
            Keyword arguments which are passed to the engine. Options 'wfheights' (for HR) and 'wfaltitudes' (for DO)
            and 'wfwidths' (for both) will be ignored as they are controlled by box_boundaries, box_midpoints and
            wf_widths.
        """
        self.geometry = geometry
        self.atmosphere = atmosphere
        self.amf_species = amf_species
        self.options = options
        self.box_boundaries = box_boundaries
        self._box_midpoints = box_midpoints
        self.wf_widths = wf_widths
        self.total_sum = total_sum
        self.total_remove = total_remove
        if not hasattr(self, '_wf_obj'):
            # doas.wf.WeightingFunctionBase object - should be set by __init__ of derived classes
            self._wf_obj = None

    @property
    def geometry(self):
        if self._wf_obj is None:
            raise NotImplementedError()
        else:
            return self._wf_obj.geometry

    @geometry.setter
    def geometry(self, value: sk.Geometry):
        if self._wf_obj is None:
            raise NotImplementedError()
        else:
            self._wf_obj.geometry = value

    @property
    def atmosphere(self):
        if self._wf_obj is None:
            raise NotImplementedError()
        else:
            return self._wf_obj.atmosphere

    @atmosphere.setter
    def atmosphere(self, value: sk.Atmosphere):
        if self._wf_obj is None:
            raise NotImplementedError()
        else:
            self._wf_obj.atmosphere = value

    @property
    def amf_species(self):
        if self._wf_obj is None:
            raise NotImplementedError()
        else:
            return self._wf_obj.wf_species[0]

    @amf_species.setter
    def amf_species(self, value: str):
        if self._wf_obj is None:
            raise NotImplementedError()
        else:
            self._wf_obj.wf_species = value

    @property
    def box_boundaries(self):
        return self._box_boundaries

    @box_boundaries.setter
    def box_boundaries(self, value: np.array):
        self._box_boundaries = value

    @property
    def options(self):
        if self._wf_obj is None:
            raise NotImplementedError()
        else:
            return self._wf_obj.options

    @options.setter
    def options(self, value):
        if self._wf_obj is None:
            raise NotImplementedError()
        else:
            self._wf_obj.options = value

    @property
    def box_midpoints(self):
        if self._box_midpoints is None:
            return (self.box_boundaries[:-1] + self.box_boundaries[1:]) / 2
        else:
            return self._box_midpoints

    @box_midpoints.setter
    def box_midpoints(self, value: Union[None, np.array]):
        self._box_midpoints = value

    @property
    def wf_widths(self):
        if self._wf_widths is None:
            return (self.box_boundaries[1:] - self.box_boundaries[:-1]) / 2
        else:
            return self._wf_widths

    @wf_widths.setter
    def wf_widths(self, value):
        self._wf_widths = value

    def _set_boxes(self):
        if len(self.box_boundaries) - len(self.box_midpoints) != 1:
            raise ValueError('Length of box_boundaries must be 1 greater than length of box_midpoints')
        if np.min(self.box_midpoints - self.box_boundaries[:-1]) < 0.0:
            raise ValueError('box_midpoints must be within the corresponding layers defined by box_boundaries')
        if np.min(self.box_boundaries[1:] - self.box_midpoints) < 0.0:
            raise ValueError('box_midpoints must be within the corresponding layers defined by box_boundaries')
        if len(self.box_midpoints) != len(self.wf_widths):
            raise ValueError('box_midpoints and wf_widths must have the same length')

        self.options['wfwidths'] = self.wf_widths
        # wfheights/wfaltitudes must be set in the engine-specific classes because the key is different

    @property
    def lines_of_sight(self):
        """
        Lines of sight at which the radiances, weighting functions, number densities and
        cross sections are specified at.

        Returns
        -------
        np.ndarray
            Array of sk.LineOfSight objects.
        """
        return self.geometry.lines_of_sight

    @property
    def reference_points(self):
        """
        Reference points corresponding to lines_of_sight

        Returns
        -------
        np.ndarray
            Reference points, described by latitude, longitude, altitude, and mjd (in that order) for each line of
            sight. Shape is '(len(lines_of_sight), 4)'. If there is a single reference point for all lines of sight,
            shape is '(1, 4)'.
        """
        raise NotImplementedError()

    @property
    def surface_height(self):
        """
        Returns
        -------
        float
            Ground altitude in metres.
        """
        if self._wf_obj is None:
            raise NotImplementedError()
        else:
            return self._wf_obj.surface_height

    @property
    def top_of_atmosphere(self):
        """
        Returns
        -------
        float
            Top of atmosphere altitude in metres.
        """
        if self._wf_obj is None:
            raise NotImplementedError()
        else:
            return self._wf_obj.top_of_atmosphere

    @property
    def atmosphere_grid(self):
        """
        Returns
        -------
        np.array
            The altitudes in metres that the atmosphere is defined at internally.
        """
        if self._wf_obj is None:
            raise NotImplementedError()
        else:
            return self._wf_obj.atmosphere_grid

    @property
    def number_densities(self):
        """
        Returns the retrieval species number densities used for the AMF calculations.

        Returns
        -------
        np.npdarray
            Number densities. Shape is (len(lines_of_sight), len(altitudes)).
            If all LOSs share a reference point, shape is (1, len(altitudes)).
        """
        rp = self.reference_points
        name = self.amf_species
        alts = self.box_midpoints

        nlos = np.size(rp, 0)
        nbox = len(alts)

        nd = np.zeros((nlos, nbox))
        for i in range(nlos):
            spec = self.atmosphere[name].species
            clim = self.atmosphere[name].climatology
            nd[i, :] = clim.get_parameter(spec, rp[i, 0], rp[i, 1], alts, rp[i, 3])
        return nd

    def cross_sections(self, wl: List[float]):
        """
        Returns the absorption cross sections used for the AMF calculations.

        Parameters
        ----------
        wl : array_like
            Wavelengths to calculate AMF's at.

        Returns
        -------
        np.npdarray
            Absorption cross sections. Shape is (len(wl), len(lines_of_sight), len(altitudes)).
            If all LOSs share a reference point, shape is (len(wl), 1, len(altitudes)).
        """
        rp = self.reference_points
        spec = self.amf_species
        alts = self.box_midpoints

        nwl = len(wl)
        nlos = np.size(rp, 0)
        nbox = len(alts)

        cs = np.zeros((nwl, nlos, nbox))

        for i in range(nlos):
            for j in range(nbox):
                clim = self.atmosphere.atmospheric_state if self.atmosphere.atmospheric_state is not None else sk.MSIS90()
                op = self.atmosphere[spec].optical_property
                cs[:, i, j] = op.calculate_cross_sections(clim, rp[i, 0], rp[i, 1], alts[j], rp[i, 3], wl).total
        return cs

    @property
    def vertical_column_density(self):
        """
        Returns the vertical column density (VCD) for amf_species over the extent of box_boundaries.
        :return:
        """
        # get all necessary values
        box_bounds = 1e2 * self.box_boundaries          # shape (nbox+1)                units cm
        box_alts = 1e2 * self.box_midpoints             # shape (nbox)                  units cm
        top = box_bounds[-1]                            # scalar                        units cm
        bottom = box_bounds[0]                          # scalar                        units cm
        nd = self.number_densities                      # shape (nrp, nbox)             units cm^-3

        # calculate the vcd (cm^-2)
        vcd = np.trapz(nd, box_alts, axis=1)            # shape (nrp,) units cm^-2
        vcd += nd[:, 0] * (box_alts[0] - bottom)        # shape (nrp,) units cm^-2
        vcd += nd[:, -1] * (top - box_alts[-1])         # shape (nrp,) units cm^-2
        return vcd

    def vertical_optical_depth(self, wl: List[float]):
        """
        Returns the vertical optical depth for amf_species over the extent of box_boundaries.

        Returns
        -------
        np.ndarray
            Optical depth
        """
        # Get dimensions
        nrp = len(self.reference_points)  # nrp = nlos for DO, nrp = 1 for HR
        nbox = len(self.box_midpoints)

        # get all necessary values
        box_bounds = 1e2 * self.box_boundaries          # shape (nbox+1)                units cm
        box_alts = 1e2 * self.box_midpoints             # shape (nbox)                  units cm
        top = box_bounds[-1]                            # scalar                        units cm
        bottom = box_bounds[0]                          # scalar                        units cm
        nd = self.number_densities                      # shape (nrp, nbox)             units cm^-3
        abs_cs = self.cross_sections(wl)                # shape (nwl, nrp, nbox)        units cm^2

        # Calculate the vod (unitless)
        vod = np.trapz(nd.reshape(1, nrp, nbox) * abs_cs, box_alts, axis=2)           # shape (nwl, nrp)
        vod += nd[:, 0].reshape(1, nrp) * abs_cs[:, :, 0] * (box_alts[0] - bottom)    # shape (nwl, nrp)
        vod += nd[:, -1].reshape(1, nrp) * abs_cs[:, :, -1] * (top - box_alts[-1])    # shape (nwl, nrp)
        return vod

    def radiances(self, wl: List[float]):
        """
        Returns the sun-normalized radiance values used for the AMF calculations.

        Parameters
        ----------
        wl : array_like
            Wavelengths to calculate radiance values at.

        Returns
        -------
        np.ndarray
            Radiances. Shape is `(len(wl), len(lines_of_sight))`.
        """
        if self._wf_obj is None:
            raise NotImplementedError()
        else:
            return self._wf_obj.radiance(wl, False)

    def weighting_functions(self, wl: List[float]):
        """
        Returns the weighting functions used for the AMF calculations.

        Parameters
        ----------
        wl : array_like
            Wavelengths to calculate AMF's at.

        Returns
        -------
        np.npdarray
            Weighting functions. Shape is `(len(wl), len(lines_of_sight), len(altitudes))`.
        """
        self._set_boxes()
        if self._wf_obj is None:
            raise NotImplementedError()
        else:
            return self._wf_obj(wl)

    def _calc_total_amf_sum(self, wl, amfs):
        # Get dimensions
        nrp = len(self.reference_points)  # nrp = nlos for DO, nrp = 1 for HR
        nbox = len(self.box_midpoints)

        # get all necessary values
        box_bounds = 1e2 * self.box_boundaries          # shape (nbox+1)                units cm
        box_widths = box_bounds[1:] - box_bounds[:-1]   # shape (nbox)                  units cm
        nd = self.number_densities                      # shape (nrp, nbox)             units cm^-3
        vcd = self.vertical_column_density

        # Calculate the total AMFs from the box AMFs
        amfs['total_amf_sum'] = np.sum(amfs['amf'] * nd[:, :].reshape(1, nrp, nbox) * box_widths.reshape(1, 1, nbox), axis=2)  # shape (nwl, nlos) units cm^-3 * cm = cm^-2
        amfs['total_amf_sum'] /= vcd.reshape(1, nrp)  # shape (nwl, nlos) units cm^-2 / cm^-2 = none

    def _calc_total_amf_remove_absorber(self, wl, amfs):

        # make the profile include points before and after the transition
        bottom = self.box_boundaries[0]
        top = self.box_boundaries[-1]
        self._wf_obj.append_atmosphere_grid(np.array([bottom - 1, bottom + 1, top - 1, top + 1]))

        # calculate true radiance with same grid
        rad = self.radiances(wl).radiance

        # should only need one reference point: HR only has one and DO should call this one LOS at a time
        rp = self.reference_points[0, :]

        # copy the profile of the absorber
        alts = self.atmosphere_grid
        spec1 = self.atmosphere[self.amf_species]
        clim1 = spec1.climatology
        prof1 = clim1.get_parameter(spec1.species, rp[0], rp[1], alts, rp[3])

        # remove the absorber within the region where the AMF is being calculated
        prof2 = prof1.copy()
        bottom = self.box_boundaries[0]
        top = self.box_boundaries[-1]
        prof2[np.logical_and(bottom <= alts, alts <= top)] = 0.0
        clim2 = sk.ClimatologyUserDefined(alts, {spec1.species: prof2})
        spec2 = sk.Species(spec1.optical_property, clim2, spec1.species)
        self.atmosphere[self.amf_species] = spec2

        # calculate radiance
        rad_no_abs = self.radiances(wl).radiance

        # restore atmosphere and options
        self.atmosphere[self.amf_species] = spec1
        self._wf_obj.restore_atmosphere_grid()

        # calculate the total vertical optical depth
        vod = self.vertical_optical_depth(wl)

        # Calculate total AMFs by removing the absorber
        amfs['total_amf_approx'] = (np.log(rad_no_abs) - np.log(rad)) / vod

    def __call__(self, wl):
        """
        Calculates wavelength dependent AMFs from weighting functions.

        Parameters
        ----------
        wl : array_like
            Wavelengths to calculate AMFs at.

        Returns
        -------
        xr.dataset
            Air mass factors, including box-AMFs, total AMFs calculated from the box-AMFs, and total AMFs calculated
            by removing the absorber.
        """

        # Get dimensions
        nbox = len(self.box_midpoints)

        # get all necessary values
        wf = self.weighting_functions(wl)               # shape (nwl, nlos, nbox)       units rad/cm^-3
        wf_widths = 1e2 * self.wf_widths                # shape (nbox)                  units cm
        abs_cs = self.cross_sections(wl)                # shape (nwl, nrp, nbox)        units cm^2

        # Convert wf = dI/dn to amf = -d(lnI)/d(tau) = -1/I dI/d(tau) where tau = n * sigma * width
        wf_species = [wfs for wfs in wf.data_vars.keys() if 'wf_' in wfs]  # wf_no2, etc.
        if len(wf_species) > 1:
            raise ValueError(f'There should only be one wf data variable. Found {len(wf_species)}: {wf_species}')

        amfs = wf.rename({wf_species[0]: 'amf'})
        amfs['amf'] = -amfs['amf'] / amfs.radiance / (abs_cs * wf_widths.reshape(1, 1, nbox))  # shape (nwl, nlos, nbox) units (rad/cm^-3)/rad/cm^2/cm = none

        # calculate the total AMFs
        if self.total_sum:
            self._calc_total_amf_sum(wl, amfs)
        if self.total_remove:
            self._calc_total_amf_remove_absorber(wl, amfs)

        return amfs


class AirMassFactorFiniteDifferenceBase(AirMassFactorBase):
    """
    Base class to access properties specific to the finite difference weighting function classes.

    Parameters
    ----------
    geometry : sk.Geometry
        Sasktran geometry object.
    atmosphere : sk.Atmosphere
        Sasktran atmosphere object.
    retrieval_species : str
        One or more species to calculate weighting functions for. Strings should match the name property of
        sk.Species objects in atmosphere.
    box_boundaries : np.array
        Array of altitudes in metres defining the boundaries of the layers the box-AMFs will be calculated for.
        Default 1 km boxes from 0 to 100 km.
    box_midpoints : None, np.array
        Array of altitudes in metres defining the point within each layer the box-AMF will be calculated at.
        Defaults to the midpoint of the layer when box_boundaries is None. Must have a length 1 less than
        box_boundaries.
    wf_widths : None, np.array
        The width in metres used to calculate the weighting functions used to calculate the box-AMFs. The width must
        be smaller than the distance between the lowest box_midpoint and the surface_height and smaller than the highest
        midpoint and the top-of-atmosphere in order to avoid unwanted interpolation effects, but it must be large
        enough to avoid numerical instability. Defaults to half the width of each layer when wf_widths is None. Must
        have the same length as box_midpoints.
    total_sum : False, bool
        True to calculate the total AMF by taking an average of the box AMFs weighted by the box vertical column.
        Requires no additional radiance calculations.
    total_remove : False, bool
        True to estimate the total AMF by calculating the radiance after removing the absorber. Only accurate
        for optically thin species. Requires two additional radiance calculations.
    method : str
        Numerical derivative method. Should be one of 'forward', 'backward' or 'central'. Central will take twice
        as long. Default 'forward'.
    step : float, List[float]
        The step size to use for the perturbation used to calculate the weighting function. A single value will
        be used at all altitudes, and a list should contain values for each altitude in box_midpoints. Default 0.1.
    relative_step : bool
        If True, step will be interpreted as a fraction and it will be multiplied by the base value to determine
        the perturbation size. If False, the value of step will be used without multiplication. Default True.
    minimum_step : float
        Minimum step size to prevent perturbations from being too small when relative_step is True. Default 1e8.
    progress_bar : bool
        True to display a progress bar during the calculation. Default False.
    **options
        Keyword arguments which are passed to the engine. Options 'wfheights' (for HR) and 'wfaltitudes' (for DO)
        and 'wfwidths' (for both) will be ignored as they are controlled by box_boundaries, box_midpoints and
        wf_widths.
    """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, amf_species: Union[str, List[str]],
                 box_boundaries: np.array=np.linspace(0, 1e5, 101), box_midpoints: Union[None, np.array]=None,
                 wf_widths: Union[None, np.array]=None, total_sum: bool=False, total_remove: bool = False,
                 method: str='forward', step: Union[float, List[float]]=0.1, relative_step: bool=True,
                 minimum_step: float=1e8, progress_bar: Union[bool, tqdm]=False, **options):
        super().__init__(geometry, atmosphere, amf_species, box_boundaries, box_midpoints, wf_widths, total_sum,
                         total_remove, **options)
        self.method = method
        self.step = step
        self.relative_step = relative_step
        self.minimum_step = minimum_step
        self.progress_bar = progress_bar
        if not hasattr(self, '_wf_obj'):
            raise NotImplementedError()

    @property
    def method(self):
        return self._wf_obj.method

    @method.setter
    def method(self, value):
        self._wf_obj.method = value

    @property
    def step(self):
        return self._wf_obj.step

    @step.setter
    def step(self, value):
        self._wf_obj.step = value

    @property
    def relative_step(self):
        return self._wf_obj.relative_step

    @relative_step.setter
    def relative_step(self, value):
        self._wf_obj.relative_step = value

    @property
    def minimum_step(self):
        return self._wf_obj.minimum_step

    @minimum_step.setter
    def minimum_step(self, value):
        self._wf_obj.minimum_step = value

    @property
    def progress_bar(self):
        return self._wf_obj.progress_bar

    @progress_bar.setter
    def progress_bar(self, value):
        self._wf_obj.progress_bar = value