import numpy as np
import sasktran as sk
from tqdm import tqdm
import doas.amf.amf
import doas.wf.disco_wf

from typing import Union, List


class AirMassFactorBaseDiscO(doas.amf.amf.AirMassFactorBase):
    """ Base class for AMF classes that use the Sasktran Discrete-Ordinates engine. """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, amf_species: str,
                 box_boundaries: np.array=np.linspace(0, 1e5, 101), box_midpoints: Union[None, np.array]=None,
                 wf_widths: Union[None, np.array]=None, total_sum: bool=False, total_remove: bool=False, **options):
        """ For parameter definitions see AirMassFactorWFBase.__init__. """
        super().__init__(geometry, atmosphere, amf_species, box_boundaries, box_midpoints, wf_widths, total_sum,
                         total_remove, **options)

    def _set_boxes(self):
        super()._set_boxes()
        self.options['wfaltitudes'] = self.box_midpoints

    @property
    def reference_points(self):
        # calculate surface_height intersections TODO add ref point customization and/or altitude control
        los = self.lines_of_sight

        ref_pts = np.zeros((len(los), 4))
        for i, s in enumerate(los):
            ref = s.ground_intersection()
            if ref is None:
                ref = s.tangent_location()
            ref_pts[i, :] = [ref.latitude, ref.longitude, ref.altitude, s.mjd]
        return ref_pts


class AirMassFactorDiscO(AirMassFactorBaseDiscO):
    """ Calculate AMFs using the Sasktran Discrete-Ordinates engine's built in weighting function algorithm. """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, amf_species: str,
                 box_boundaries: np.array=np.linspace(0, 1e5, 101), box_midpoints: Union[None, np.array]=None,
                 wf_widths: Union[None, np.array]=None, total_sum: bool=False, total_remove: bool=False, **options):
        """ For parameter definitions see Ai rMassFactorWFBase.__init__. """
        self._wf_obj = doas.wf.disco_wf.WeightingFunctionDiscO(geometry, atmosphere, amf_species, **options)
        super().__init__(geometry, atmosphere, amf_species, box_boundaries, box_midpoints, wf_widths, total_sum,
                         total_remove, **options)


class AirMassFactorFiniteDifferenceDiscO(AirMassFactorBaseDiscO, doas.amf.amf.AirMassFactorFiniteDifferenceBase):
    """ Calculate AMFs using the Sasktran Discrete-Ordinates engine to get finite-difference weighting functions. """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, amf_species: str,
                 box_boundaries: np.array=np.linspace(0, 1e5, 101), box_midpoints: Union[None, np.array]=None,
                 wf_widths: Union[None, np.array]=None, total_sum: bool=False, total_remove: bool=False,
                 method: str='forward', step: Union[float, List[float]]=0.1, relative_step: bool=True,
                 minimum_step: float=1e8, progress_bar: Union[bool, tqdm]=False, **options):
        """ For parameter definitions see AirMassFactorFiniteDifferenceBase.__init__. """
        self._wf_obj = doas.wf.disco_wf.WeightingFunctionFiniteDifferenceDiscO(
            geometry, atmosphere, amf_species, method, step, relative_step, minimum_step, progress_bar, **options
        )
        AirMassFactorBaseDiscO.__init__(
            self, geometry, atmosphere, amf_species, box_boundaries, box_midpoints, wf_widths, total_sum, total_remove,
            **options
        )
        doas.amf.amf.AirMassFactorFiniteDifferenceBase.__init__(
            self, geometry, atmosphere, amf_species, box_boundaries, box_midpoints, wf_widths, total_sum, total_remove,
            method, step, relative_step, minimum_step, progress_bar, **options
        )
