import numpy as np
import sasktran as sk
from tqdm import tqdm
import doas.amf.amf
import doas.wf.finite_difference_wf
import doas.wf.hr_wf

from typing import List, Union


class AirMassFactorBaseHR(doas.amf.amf.AirMassFactorBase):
    """ Base class for AMF classes that use the Sasktran High-Resolution engine. """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, amf_species: str,
                 box_boundaries: np.array=np.linspace(0, 1e5, 101), box_midpoints: Union[None, np.array]=None,
                 wf_widths: Union[None, np.array]=None, total_sum: bool=False, total_remove: bool=False, **options):
        """ For parameter definitions see AirMassFactorWFBase.__init__. """
        super().__init__(geometry, atmosphere, amf_species, box_boundaries, box_midpoints, wf_widths, total_sum,
                         total_remove, **options)

    def _set_boxes(self):
        super()._set_boxes()
        self.options['wfheights'] = self.box_midpoints

    @property
    def reference_points(self):
        # fetch reference point
        ref_pts = np.zeros((1, 4))
        engine = sk.EngineHR(self.geometry, self.atmosphere, [500.0], self.options)
        engine._add_lines_of_sight()
        engine._initialize_model()
        ref_pts[0, :] = engine.model_parameters['referencepoint']
        return ref_pts


class AirMassFactorHR(AirMassFactorBaseHR):
    """ Calculate AMFs using the Sasktran High-Resolution engine's built in weighting function algorithm. """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, amf_species: str,
                 box_boundaries: np.array=np.linspace(0, 1e5, 101), box_midpoints: Union[None, np.array]=None,
                 wf_widths: Union[None, np.array]=None, total_sum: bool=False, total_remove: bool=False, **options):
        """ For parameter definitions see AirMassFactorWFBase.__init__. """
        self._wf_obj = doas.wf.hr_wf.WeightingFunctionHR(geometry, atmosphere, amf_species, **options)
        super().__init__(geometry, atmosphere, amf_species, box_boundaries, box_midpoints, wf_widths, total_sum,
                         total_remove, **options)


class AirMassFactorFiniteDifferenceHR(AirMassFactorBaseHR, doas.amf.amf.AirMassFactorFiniteDifferenceBase):
    """ Calculate AMFs using the Sasktran High-Resolution engine to get finite-difference weighting functions. """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, amf_species: str,
                 box_boundaries: np.array=np.linspace(0, 1e5, 101), box_midpoints: Union[None, np.array]=None,
                 wf_widths: Union[None, np.array]=None, total_sum: bool=False, total_remove: bool=False,
                 method: str='forward', step: Union[float, List[float]]=0.1, relative_step: bool=True,
                 minimum_step: float=1e8, progress_bar: Union[bool, tqdm]=False, **options):
        """ For parameter definitions see AirMassFactorFiniteDifferenceBase.__init__. """
        self._wf_obj = doas.wf.hr_wf.WeightingFunctionFiniteDifferenceHR(
            geometry, atmosphere, amf_species, method, step, relative_step, minimum_step, progress_bar, **options
        )
        AirMassFactorBaseHR.__init__(
            self, geometry, atmosphere, amf_species, box_boundaries, box_midpoints, wf_widths, total_sum, total_remove,
            **options
        )
        doas.amf.amf.AirMassFactorFiniteDifferenceBase.__init__(
            self, geometry, atmosphere, amf_species, box_boundaries, box_midpoints, wf_widths, total_sum, total_remove,
            method, step, relative_step, minimum_step, progress_bar, **options
        )