import numpy as np
import matplotlib.pyplot as plt
import doas.ring.rrs_cross_sections


class RingSpectrum(object):
    """
    Calculates the ring spectrum from rotational Raman
    scattering of O2 and N2 particles. For use in DOAS retrievals.

    Parameters
    ----------
    solar_spectrum_wavelength : np.ndarray
        Array of wavelengths corresponding to the irradiance data.
    solar_spectrum_irradiance : np.ndarray
        Solar top-of-atmosphere irradiance data.

    Properties
    ----------

    plot : bool
        Indicates whether to produce a plot showing the stages of the
        calculations. Defaults to False.
    n2_levels : int
        Number of N2 energy levels to consider in the calculation.
        Defaults to 25.
    o2_levels : int
        Number of O2 energy levels to consider in the calculation.
        Defaults to 25.
    temperature : float
        Effective temperature. Defaults to 250K.
    full : bool
        True to calculate the true RRS cross sections at each individual wavelength for convolution. False to
        use the cross sections at one wavelength for the whole convolution.
    polynomial_order : int
        The order of the polynomial to remove from the final spectrum. Defaults to 3.
    polynomial_removal : str
    The method of removing the polynomial. Should be 'subtract', 'divide' or None.

    """
    def __init__(self, solar_spectrum_wavelength, solar_spectrum_irradiance):
        self._wl = solar_spectrum_wavelength
        self._irr = solar_spectrum_irradiance

        self.plot = False
        self.n2_levels = 25
        self.o2_levels = 25
        self.temperature = 250.0
        self.full = False
        self.polynomial_order = 3
        self.polynomial_removal = 'subtract'

    @property
    def plot(self):
        return self._plot

    @property
    def n2_levels(self):
        return self._n2_levels

    @property
    def o2_levels(self):
        return self._o2_levels

    @property
    def temperature(self):
        return self._temp

    @property
    def full(self):
        return self._full

    @property
    def polynomial_order(self):
        return self._poly_order

    @property
    def polynomial_removal(self):
        return self._poly_remove

    @plot.setter
    def plot(self, value):
        self._plot = value

    @n2_levels.setter
    def n2_levels(self, value):
        self._n2_levels = value

    @o2_levels.setter
    def o2_levels(self, value):
        self._o2_levels = value

    @temperature.setter
    def temperature(self, value):
        self._temp = value

    @full.setter
    def full(self, value):
        self._full = value

    @polynomial_order.setter
    def polynomial_order(self, value):
        self._poly_order = value

    @polynomial_removal.setter
    def polynomial_removal(self, value):
        if isinstance(value, str):
            if value.lower() == 'subtract':
                self._poly_remove = 'subtract'
            elif value.lower() == 'divide':
                self._poly_remove = 'divide'
            else:
                self._poly_remove = None
        else:
            self._poly_remove = None

    def get_ring_spectrum(self, wavelength):
        """
        Calculates the ring spectrum from rotational Raman
        scattering of O2 and N2 particles. For use in DOAS retrievals.

        Parameters
        ----------
        wavelength : np.ndarray
            The wavelengths for the ring spectrum calculation. If these
            wavelengths do not fall within the wavelength range of the
            irradiance data, results will be weird.

        Returns
        -------
        np.ndarray
            Array with the ring spectrum.
        """

        ring = np.zeros_like(wavelength)
        for i, wl in enumerate(wavelength):
            rrso2 = doas.ring.rrs_cross_sections.RRSCrossSectionsO2(wl)
            rrsn2 = doas.ring.rrs_cross_sections.RRSCrossSectionsN2(wl)
            wlo2, cso2 = rrso2.calc_cross_sections(self._o2_levels, self._temp)
            wln2, csn2 = rrsn2.calc_cross_sections(self._n2_levels, self._temp)
            if len(wlo2) > len(wln2):
                wlh = wlo2
                csn2 = np.interp(wlh, wln2, csn2, left=0.0, right=0.0)
            else:
                wlh = wln2
                cso2 = np.interp(wlh, wlo2, cso2, left=0.0, right=0.0)
            csatmo = 0.21 * cso2 + 0.78 * csn2
            irrh = np.interp(wlh, self._wl, self._irr)
            ring[i] = np.convolve(irrh, csatmo, 'valid')[0]

        return ring

    def get_differential_ring_spectrum(self, wavelength):
        """
        Calculates the differential ring spectrum from rotational Raman
        scattering of O2 and N2 particles. For use in DOAS retrievals.

        Parameters
        ----------
        wavelength : np.ndarray
            The wavelengths for the ring spectrum calculation. If these
            wavelengths do not fall within the wavelength range of the
            irradiance data, results will be weird.

        Returns
        -------
        np.ndarray
            Array with the differential ring spectrum.
        """

        if self._plot is True:
            plt.subplot(221)
            plt.plot(self._wl, self._irr)
            plt.xlabel('Wavelength (nm)')
            plt.ylabel('Irradiance (photons/s.nm.cm$^2$)')
            plt.title('Top of Atmosphere Solar Spectrum')

        # get the ring effect cross sections
        centre_wavelength = 0.5 * (wavelength[0] + wavelength[-1])
        rrso2 = doas.ring.rrs_cross_sections.RRSCrossSectionsO2(centre_wavelength)
        rrsn2 = doas.ring.rrs_cross_sections.RRSCrossSectionsN2(centre_wavelength)
        wlo2, cso2 = rrso2.calc_cross_sections(self._o2_levels, self._temp)
        wln2, csn2 = rrsn2.calc_cross_sections(self._n2_levels, self._temp)

        # get the width of the ring effect
        width = max([wlo2[-1] - wlo2[0], wln2[-1] - wln2[0]])
        extend = round(width / 2, 3) + 0.001

        # zero-pad the cross-sections to match the window
        wl_fine = np.arange(wavelength[0] - extend,
                            wavelength[-1] + extend + 0.001, 0.001)
        csn2 = np.interp(wl_fine, wln2, csn2, left=0.0, right=0.0)
        cso2 = np.interp(wl_fine, wlo2, cso2, left=0.0, right=0.0)
        csatmo = 0.21 * cso2 + 0.78 * csn2

        # perform the convolution
        if self.full:
            conv = self.get_ring_spectrum(wavelength)
            diff = conv / np.interp(wavelength, self._wl, self._irr)
        else:
            irr_fine = np.interp(wl_fine, self._wl, self._irr)
            conv = np.convolve(irr_fine, csatmo, 'same')
            conv = np.interp(wavelength, wl_fine, conv)
            # divide out the original solar spectrum to get the differential spectrum
            diff = conv / np.interp(wavelength, wl_fine, irr_fine)

        if self._plot is True:
            plt.subplot(222)
            plt.plot(wl_fine, 0.21 * cso2, label='O2')
            plt.plot(wl_fine, 0.78 * csn2, label='N2')
            plt.xlabel('Wavelength (nm)')
            plt.ylabel('Cross Section (cm$^2$)')
            plt.title('RRS Cross Sections at Center Wavelength')
            plt.legend()

        # fit a cubic polynomial to remove
        if self.polynomial_removal is 'subtract':
            p = np.polyfit(wavelength, diff, self.polynomial_order)
            diff_no_poly = diff - np.polyval(p, wavelength)
        elif self.polynomial_removal is 'divide':
            p = np.polyfit(wavelength, diff, self.polynomial_order)
            diff_no_poly = diff / np.polyval(p, wavelength)
        else:
            diff_no_poly = diff

        if self._plot is True:
            plt.subplot(223)
            plt.plot(wavelength, conv)
            plt.xlabel('Wavelength (nm)')
            plt.ylabel('Ring Effect Source Spectrum (photons.cm$^2$)')
            plt.title('RRS Cross Sections Convolved with Solar Spectrum')

            plt.subplot(224)
            plt.plot(wavelength, diff_no_poly)
            plt.xlabel('Wavelength (nm)')
            plt.ylabel('Differential Ring Spectrum')
            plt.title('Differential Ring Spectrum')

            plt.show()

        return diff_no_poly
