import numpy as np
from typing import MutableMapping, List


class Equation:
    """
    DOAS Equation base class. Equation objects are callable, providing the forward calculation used to fit parameters
    such as absorber slant column densities and low-order polynomial coefficients to a spectrum.
    """
    def __init__(self, poly_degree: int, **kwargs):
        """
        Parameters
        ----------
        poly_degree : int
            Degree of slow spectrally varying polynomial.
        kwargs
            Additional attributes which will be added the the object.
        """
        self.poly_degree = poly_degree
        self.__dict__.update(kwargs)

    def parse_fit_params(self, *fit_params: List[float]) -> MutableMapping[str, np.ndarray]:
        """
        Parse the fitted equation parameters to a OrderedDict.

        Parameters
        ----------
        *fit_params : array_like
            Fitted equation parameters.

        Returns
        -------
        collections.OrderedDict
            Ordered dictionary containing numpy arrays of grouped equation parameters. Guaranteed to contain 'slant_cd'
            and 'poly_coeffs'.
        """
        raise NotImplementedError()

    def unparse_fit_params(self, fit_params: MutableMapping[str, np.ndarray]) -> List[float]:
        """
        Performs the opposite function as parse_fit_params.

        Parameters
        ----------
        fit_params : collections.OrderedDict
            Ordered dictionary containing numpy arrays of grouped equation parameters. Output from parse_fit_params.

        Returns
        -------
        list
            A list of equation parameters suitable for passing to __call__.
        """
        raise NotImplementedError()


class LinearEquation(Equation):
    """
    DOAS Equation base class. Equation objects are callable, providing the forward calculation used to fit parameters
    such as absorber slant column densities and low-order polynomial coefficients to a spectrum.
    """
    def __init__(self, poly_degree: int, **kwargs):
        super().__init__(poly_degree, **kwargs)
        raise NotImplementedError()


class NonLinearEquation(Equation):
    def __init__(self, poly_degree: int, **kwargs):
        super().__init__(poly_degree, **kwargs)
        self._jac_defined = getattr(self, '_jac_defined', False)
        self._cov_defined = getattr(self, '_cov_defined', False)

    def __call__(self, wl: np.ndarray, *fit_params: List[float]) -> np.ndarray:
        """ Model the slant radiance.

        Parameters
        ----------
        wl : np.ndarray
            Wavelengths to calculate modelled radiances at.
        *fit_params : List[float]
            The equations parameters.

        Returns
        -------
        np.ndarray
            The modelled radiances (corresponding to `wl`).
        """
        raise NotImplementedError()

    @property
    def jacobian_is_defined(self):
        """bool: True if the equation defines a jacobian, False otherwise."""
        return self._jac_defined

    @property
    def covariance_is_defined(self):
        """bool: True if the equation defines a covariance matrix, False otherwise."""
        return self._cov_defined

    @property
    def num_fit_params(self):
        """ Expected number of fit parameters. """
        raise NotImplementedError()

    def jacobian(self, wl: np.ndarray, *fit_params) -> np.ndarray:
        """ Jacobian of modelled slant radiance. It has shape (len(wl), len(fit_params)). """
        raise NotImplementedError()

    def covariance(self, wl: np.ndarray, rad: np.ndarray, sigma: np.ndarray, *fit_params) -> np.ndarray:
        """ Calculate the covariance matrix of the least squares fit. It has shape (len(fit_params), len(fit_params).
        If this is not provided, fit_spectra will use the covariance matrix estimated by scipy.optimize.curve_fit. """
        raise NotImplementedError()
