from typing import MutableMapping
import numpy as np
import scipy.optimize as spo

from doas.scd.equation import Equation, LinearEquation, NonLinearEquation


def fit_spectra(wl: np.ndarray, rad: np.ndarray, eqn: Equation, **kwargs):
    """ Fit measurements to a DOAS equation.

    Parameters
    ----------
    wl : np.ndarray
        Measurement wavelengths.
    rad : np.ndarray
        Measured radiances.
    eqn : doas.Equation
        DOAS equation to be fit.

    Returns
    -------
    MutableMapping[str, np.ndarray]
        Mapping of the fit equation parameters. See doas.Equation.parse_fit_params.
    """
    if issubclass(type(eqn), LinearEquation):
        return _fit_spectra_linear(wl, rad, eqn, **kwargs)
    elif issubclass(type(eqn), NonLinearEquation):
        return _fit_spectra_nonlinear(wl, rad, eqn, **kwargs)
    else:
        raise TypeError("Unknown DOAS equation type!")


def _fit_spectra_linear(wl: np.ndarray, rad: np.ndarray, eqn: LinearEquation, **kwargs):
    raise NotImplementedError()


def _fit_spectra_nonlinear(wl: np.ndarray, rad: np.ndarray, eqn: NonLinearEquation,
                           return_one_sigma: bool=False,
                           return_cov: bool=False,
                           return_chi2: bool=False,
                           **kwargs):
    """ Fit measurements to a non-linear DOAS equation.

    Parameters
    ----------
    wl : np.ndarray
        See doas.fit_spectra.
    rad : np.ndarray
        See doas.fit_spectra.
    eqn : doas.Equation
        See doas.fit_spectra.
    return_one_sigma : bool, False
        If True the one sigma values will be returned.
    return_cov : bool, False
        If True the covariance matrix will be returned.
    return_chi2 : bool, False
        If True the chi squared values will be returned.
    kwargs
        Keyword arguments will be passed to scipy.optimize.curve_fit when fitting.

    Returns
    -------
    MutableMapping[str, np.ndarray]
        Mapping of the fit equation parameters. See doas.Equation.parse_fit_params.
    MutableMapping[str, np.ndarray]
        One sigma values if return_one_sigma was True.
    np.ndarray
        Covariance matrix if return_cov was True.
    np.ndarray
        Chi squared values if return_chi2 was True.

    Examples
    --------
    >>> import doas
    >>> import sasktran as sk
    >>> import numpy as np
    >>> from scipy.interpolate import interp1d
    >>> # set up geometry
    >>> geometry = sk.geometry.NadirGeometry()
    >>> lat, lon, mjd = 52.1334, -106.6314, 57906.7
    >>> uofs = sk.Geodetic()
    >>> uofs.from_lat_lon_alt(lat, lon, 0.0)
    >>> tempo = sk.Geodetic()
    >>> tempo.from_lat_lon_alt(0.0, -100.0, 3.5786e7)
    >>> geometry.from_lat_lon(mjd, tempo, uofs)
    >>> geometry.reference_point = [lat, lon, 0.0, mjd]
    >>> # set up atmosphere
    >>> atmosphere = sk.Atmosphere()
    >>> atmosphere['air'] = sk.Species(sk.Rayleigh(), sk.MSIS90(), 'SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3')
    >>> atmosphere['no2'] = sk.Species(sk.NO2Vandaele1998(), sk.Pratmo(), 'SKCLIMATOLOGY_NO2_CM3')
    >>> atmosphere['o3'] = sk.Species(sk.O3DBM(), sk.Labow(), 'SKCLIMATOLOGY_O3_CM3')
    >>> # calculate simulated radiance
    >>> wavelengths = np.arange(405, 465)
    >>> engine = sk.EngineHR(geometry, atmosphere, wavelengths)
    >>> radiance = engine.calculate_radiance().flatten()
    >>> # fetch cross sections
    >>> optical_property = sk.NO2Vandaele1998()
    >>> cross_sections = optical_property.calculate_cross_sections(sk.MSIS90(), lat, lon, 2.7e4, mjd, wavelengths)
    >>> cross_sections_callable = interp1d(wavelengths, cross_sections.total)
    >>> # set up DOAS equation
    >>> equation = doas.OMNO2(poly_degree=5, cross_sections=[cross_sections_callable], ring_spectra=None, solar_irradiance=None)
    >>> # fit the equation to the data
    >>> fit_params, one_sigma = doas.fit_spectra(wavelengths, radiance, equation, return_one_sigma=True)
    >>> # print the no2 scd and its 1-sigma error
    >>> print(fit_params['slant_cd'])
    [  1.11720265e+16]
    >>> print(one_sigma['slant_cd'])
    [  6.88881542e+14]
    """
    jac = eqn.jacobian if eqn.jacobian_is_defined else None

    # set length of fit_params if no guess is provided
    p0 = kwargs.pop('p0', np.ones(eqn.num_fit_params))

    # do the non linear fit
    fit_params, cov = spo.curve_fit(eqn, wl, rad, jac=jac, p0=p0, **kwargs)

    # calculate the chi squared value
    sigma = kwargs.get('sigma', np.ones_like(wl))  # curve_fit defaults to 1s
    chi2 = np.sum(((rad - eqn(wl, *fit_params)) / sigma) ** 2) / (len(rad) - len(fit_params))

    # get the analytic covariance, if it is defined
    if eqn.covariance_is_defined:
        cov = eqn.covariance(wl, rad, sigma, *fit_params)
        if not kwargs.get('absolute_sigma', False):
            cov *= chi2 / (len(wl) - len(fit_params))

    # organize the results
    results = eqn.parse_fit_params(*fit_params)

    # build return values based on what was asked for
    rv = [results]
    if return_one_sigma:
        one_sigma = eqn.parse_fit_params(*[np.sqrt(d) if d > 0 else np.nan for d in np.diag(cov)])
        rv.append(one_sigma)
    if return_cov:
        rv.append(cov)
    if return_chi2:
        rv.append(chi2)
    if len(rv) == 1:
        rv = rv[0]
    return rv
