from typing import Callable, List, MutableMapping
import numpy as np
from doas.scd.equation import NonLinearEquation
from scipy.misc import derivative
from scipy.interpolate import interp1d


class OMNO2(NonLinearEquation):
    """
    DOAS OMNO2 equation. It is designed to be passed to fit_spectra().

    This class implements a non-linear DOAS equation suitable for use in the OMNO2 algorithm. The equation consists
    of a polynomial term to account for slow-changing spectral features, absorption from a list of absorbers, a
    ring spectrum term and a wavelength shift correction.

    The equation depends on the following parameters, in addition to wavelength:
     - slant_cd (slant column density): slant column densities in molecules/cm2 for each species in cross_sections
     - poly_coeffs (polynomial coefficients): poly_degree + 1 coefficients starting with the highest order
     - ring_coeff (ring spectrum coefficient): the amplitude of the ring effect (only include if ring_spectrum is
     supplied)
     - wavlen_shift (wavelength shift): the shift in nm between a measured earth radiance and a reference solar
     irradiance (only include if solar_irradiance is supplied)
    These parameters must be in this order when passed methods that accept *fit_params.
    """
    def __init__(self, poly_degree: int=5, cross_sections=None, ring_spectra=None, solar_irradiance=None, **kwargs):
        """
        Parameters
        ----------
        poly_degree : int, 5
            Degree of slow spectrally varying model polynomial.
        cross_sections : List[Callable[[np.ndarray], np.ndarray]], None
            See the `cross_sections` property.
        ring_spectra : Callable[[np.ndarray], np.ndarray], None
            See the `ring_spectra` property.
        solar_irradiance : Callable[[np.ndarray], np.ndarray], None
            See the `solar_irradiance` property.
        **kwargs
            Passed to base class.
        """
        super().__init__(poly_degree, **kwargs)
        self._jac_defined = True
        self._cov_defined = False  # There are numerical issues inverting the matrix, scipy's estimate is more reliable

        self.cross_sections = cross_sections
        self.ring_spectra = ring_spectra
        self.solar_irradiance = solar_irradiance

        self._rad = None
        self._last_params = None

    @property
    def cross_sections(self):
        """List[Callable[[np.ndarray], np.ndarray]]: Interpolations of the absorption cross sections for each species
        to be considered. If `cross_sections` is None then absorption will be neglected.
        """
        return self._cross_sections

    @cross_sections.setter
    def cross_sections(self, cs: List[Callable[[np.ndarray], np.ndarray]]):
        if cs is None:
            cs = []
        self._cross_sections = cs

    @property
    def ring_spectra(self):
        """Callable[[np.ndarray], np.ndarray]: An interpolation of the ring spectrum. If `ring_spectra` is None then
        the ring effect is neglected.
        """
        return self._ring_spectra

    @ring_spectra.setter
    def ring_spectra(self, spectra: Callable[[np.ndarray], np.ndarray]):
        self.consider_ring_effect = True
        if spectra is None:
            def spectra(w):
                return np.zeros_like(w)
            self.consider_ring_effect = False
        self._ring_spectra = spectra

    @property
    def solar_irradiance(self):
        """Callable[[np.ndarray], np.ndarray]: An interpolation of the solar irradiance. If `solar_irradiance` is None
        then the wavelength shifting correction is neglected.
        """
        return self._solar_irradiance

    @solar_irradiance.setter
    def solar_irradiance(self, irradiance: Callable[[np.ndarray], np.ndarray]):
        self.consider_shifted_wavlen = True
        if irradiance is None:

            def irradiance(w):
                return np.ones_like(w)

            def irradiance_dv(w):
                return np.zeros_like(w)
            self.consider_shifted_wavlen = False
        else:
            irradiance_dv = self._calc_solar_derivative(irradiance)
        self._solar_irradiance = irradiance
        self._solar_irradiance_dv = irradiance_dv

    @property
    def num_fit_params(self):
        return len(self.cross_sections) + self.poly_degree + 1 + \
            self.consider_ring_effect + self.consider_shifted_wavlen

    def _calc_solar_derivative(self, irradiance: interp1d):
        """
        Numerically calculate the derivative of irradiance. Cubic interpolation
        is recommended to keep the derivative smooth (linear interpolation is
        strongly discouraged).
        """
        assert isinstance(irradiance, interp1d)

        wl = irradiance.x
        dv = np.zeros_like(wl)
        # use scipy's central difference method for the middle points
        dv[1:-1] = derivative(irradiance, wl[1:-1], dx=0.001)
        # manually use forward/backward difference for the first/last points
        dv[0] = (irradiance.y[1] - irradiance.y[0]) / (wl[1] - wl[0])
        dv[-1] = (irradiance.y[-1] - irradiance.y[-2]) / (wl[-1] - wl[-2])
        return interp1d(wl, dv, kind='cubic')  # default to cubic interpolation

    def _parse_fit_params_to_tuple(self, *fit_params: List[float]) -> tuple:
        """ Parse the fit parameters to a tuple.

        The format of the parameter list is [slant_cd ...], poly_coeffs ... [, ring_coeff] [,wavlen_shift].

         Parameters
         ----------
         *fit_params : List[float]
            The equation parameters.

        Returns
        -------
        tuple
            Groupings of the equation parameters. Format is (slant_cd, poly_coeffs, ring_coeff, wavlen_shift).
        """
        fit_params = list(fit_params)

        slant_cd = None
        if len(self.cross_sections) > 0:
            num_species = len(self.cross_sections)
            slant_cd = np.array(fit_params[:num_species])
            fit_params = fit_params[num_species:]

        poly_coeffs = None
        if self.poly_degree >= 0:
            poly_coeffs = np.array(fit_params[:self.poly_degree + 1])
            fit_params = fit_params[self.poly_degree + 1:]

        ring_coeff = None
        if self.consider_ring_effect:
            ring_coeff = np.array([fit_params[0]])
            del fit_params[0]

        wavlen_shift = None
        if self.consider_shifted_wavlen:
            wavlen_shift = np.array([fit_params[0]])
            del fit_params[0]

        return slant_cd, poly_coeffs, ring_coeff, wavlen_shift

    def parse_fit_params(self, *fit_params: List[float]) -> MutableMapping[str, np.ndarray]:
        """ See doas.Equation.parse_fit_params."""
        slant_cd, poly_coeffs, ring_coeff, wavlen_shift = self._parse_fit_params_to_tuple(*fit_params)
        keys, values = [], []
        if slant_cd is not None:
            keys += ['slant_cd']
            values += [slant_cd]
        if poly_coeffs is not None:
            keys += ['poly_coeffs']
            values += [poly_coeffs]
        if ring_coeff is not None:
            keys += ['ring_coeff']
            values += [ring_coeff]
        if wavlen_shift is not None:
            keys += ['wavlen_shift']
            values += [wavlen_shift]

        return dict(zip(keys, values))

    def unparse_fit_params(self, fit_params: MutableMapping[str, np.ndarray]) -> List[float]:
        fit_list = []
        keys = ['slant_cd', 'poly_coeffs', 'ring_coeff', 'wavlen_shift']
        for key in keys:
            if key in fit_params:
                fit_list += list(fit_params[key])
        return fit_list

    def jacobian(self, wl: np.ndarray, *fit_params: List[float]) -> np.ndarray:
        """ See doas.NonLinearEquation.jacobian. """
        # If the radiance was just calculated with the same fit parameters
        # (which it will be during curve_fit), then the radiance does not
        # have to be recalculated
        if fit_params == self._last_params:
            rad = self._rad
        else:
            rad = self(wl, *fit_params)

        # parse the fit parameters
        slant_cd, poly_coeffs, ring_coeff, wavlen_shift = self._parse_fit_params_to_tuple(*fit_params)

        k = len(fit_params)
        l = len(poly_coeffs)
        m = len(self.cross_sections)
        n = len(wl)

        # evaluate polynomial at each wavelength
        poly = np.polyval(poly_coeffs, wl)

        # get cross section matrix (n, m)
        cross_sections = np.ndarray((n, m))
        for i, cs in enumerate(self.cross_sections):
            cross_sections[:, i] = cs(wl)

        jac = np.ones((n, k)) * np.reshape(rad, (n, 1))
        # derivates wrt slant column densities
        jac[:, :m] *= -cross_sections
        # derivatives wrt polynomial coefficients
        jac[:, m:m + l] *= np.reshape(wl, (n, 1)) ** np.reshape(np.arange(l, dtype=np.float)[::-1], (1, l)) / np.reshape(poly, (n, 1))
        # derivative wrt ring coefficient
        if self.consider_ring_effect:
            jac[:, m + l] *= self.ring_spectra(wl) / (1 + ring_coeff * self.ring_spectra(wl))
        # derivative wrt wavelength shift
        if self.consider_shifted_wavlen:
            jac[:, -1] *= self._solar_irradiance_dv(wl + wavlen_shift) / self.solar_irradiance(wl + wavlen_shift)

        return jac

    def covariance(self, wl: np.ndarray, rad: np.ndarray, sigma: np.ndarray, *fit_params: List[float]) -> np.ndarray:
        """ See doas.NonLinearEquation.covariance. """
        slant_cd, poly_coeffs, ring_coeff, wavlen_shift = self._parse_fit_params_to_tuple(*fit_params)
        if fit_params == self._last_params:
            doas_rad = self._rad
        else:
            doas_rad = self(wl, *fit_params)
        jac = self.jacobian(wl, *fit_params)

        n = len(wl)
        k = len(fit_params)
        m = len(self.cross_sections)
        l = len(poly_coeffs)

        # get cross section matrix (n, m) where n = num_wavelengths, m=num_species
        cross_sections = np.ndarray((n, m))
        for i, cs in enumerate(self.cross_sections):
            cross_sections[:, i] = cs(wl)

        # evaluate polynomial at each wavelength
        poly = np.polyval(poly_coeffs, wl)

        # get the ring spectrum
        if self.consider_ring_effect:
            ring = self.ring_spectra(wl)

        # get the solar spectrum and derivatives at the shifted wavelengths
        if self.consider_shifted_wavlen:
            solar = self.solar_irradiance(wl + wavlen_shift)
            dsolar = self._solar_irradiance_dv(wl + wavlen_shift)
            d2solar = self._calc_solar_derivative(self._solar_irradiance_dv)(wl + wavlen_shift)

        # set up the multipliers for each derivative - they work for all first derivatives and many second derivatives
        multiplier = np.zeros((n, k))
        multiplier[:, :m] = -cross_sections
        for i in range(l):
            multiplier[:, m + i] = wl ** (l - i - 1) / poly
        if self.consider_ring_effect:
            multiplier[:, -2] = ring / (1 + ring_coeff * ring)
        if self.consider_shifted_wavlen:
            multiplier[:, -1] = dsolar / solar

        # calculate the second derivatives
        d2 = np.ones((n, k, k)) * np.reshape(doas_rad, (n, 1, 1))
        # many just require two multipliers, so start by applying them everywhere
        d2 *= np.reshape(multiplier, (n, 1, k)) * np.reshape(multiplier, (n, k, 1))
        # overwrite the exceptions
        d2[:, m:m + l, m:m + l] = 0
        if self.consider_ring_effect:
            d2[:, m + l, m + l] *= -1
        if self.consider_shifted_wavlen:
            d2[:, -1, -1] = doas_rad * (d2solar / solar - multiplier[:, -1] ** 2)

        # second derivatives of the chi^2 function give elements of the inverse of the covariance matrix
        t1 = np.reshape(jac, (n, 1, k)) * np.reshape(jac, (n, k, 1))
        t2 = np.reshape(rad - doas_rad, (n, 1, 1)) * d2
        covinv = np.sum((t1 - t2) / np.reshape(sigma**2, (n, 1, 1)), 0)
        return np.linalg.inv(covinv)

    def __call__(self, wl: np.ndarray, *fit_params: List[float]) -> np.ndarray:
        """See doas.NonLinearEquation.__call__."""

        # parse the fit parameters
        slant_cd, poly_coeffs, ring_coeff, wavlen_shift = self._parse_fit_params_to_tuple(*fit_params)

        # evaluate polynomial at each wavelength
        if self.poly_degree >= 0:
            poly = np.polyval(poly_coeffs, wl)
        else:
            poly = np.ones_like(wl)

        # get cross section matrix (n, m) where n = num_wavelengths, m=num_species
        cross_sections = np.ndarray((len(wl), len(self.cross_sections)))
        for i, cs in enumerate(self.cross_sections):
            cross_sections[:, i] = cs(wl)

        # calculate the absorption as cross-section x slant-number-density
        if len(self.cross_sections) > 0:
            absorption = np.matmul(cross_sections, slant_cd)
        else:
            absorption = np.zeros_like(wl)

        # calculate the radiance
        rad = poly * np.exp(-absorption)
        if self.consider_ring_effect:
            rad *= 1 + ring_coeff * self.ring_spectra(wl)
        if self.consider_shifted_wavlen:
            wavlen_shift_correction = self.solar_irradiance(wl + wavlen_shift) / self.solar_irradiance(wl)
            rad *= wavlen_shift_correction

        # store the radiance and the fit parameters to speed up jacobian calls
        self._rad = rad
        self._last_params = fit_params

        return rad