import numpy as np
import sasktran as sk
import xarray as xr

from typing import List, Union
from tqdm import tqdm

import doas.wf.wf


class WeightingFunctionFiniteDifferenceBase(doas.wf.wf.WeightingFunctionBase):
    """
    Base class for weighting function calculation using a finite difference approximation.

    Parameters
    ----------
    geometry: sk.Geometry
        See doas.wf.wf.WeightingFunctionBase
    atmosphere: sk.Atmosphere
        See doas.wf.wf.WeightingFunctionBase
    method: str
        Determines which finite difference scheme is used. One of 'forward', 'backward' or 'central'
    step: Union[float, List[float]], 0.1
        The step used to perturb one layer of the atmosphere. Interpretation depends on relative_step. Should be
        a float or a list/array with length matching wf_altitudes/wf_widths.
    relative_step: bool, True
        If true, step is interpreted as a relative step. If false, it is interpreted as an absolute step.
    progress_bar: bool, False
        If true, a progress bar will be displayed.
    **options
        See doas.wf.wf.WeightingFunctionBase
    """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, wf_species: List[str]=None,
                 method: str='forward', step: Union[float, List[float]]=0.01, relative_step: bool=True,
                 minimum_step: float=1e8, progress_bar: Union[bool, tqdm]=False, **options):
        super().__init__(geometry, atmosphere, wf_species=wf_species, **options)
        self._step = step
        self._method = method
        self._rel_step = relative_step
        self._min_step = minimum_step
        self._pbar = progress_bar

    @property
    def method(self) -> str:
        return self._method

    @method.setter
    def method(self, value: str):
        self._method = value

    @property
    def step(self) -> Union[float, List[float]]:
        return self._step

    @step.setter
    def step(self, value: Union[float, List[float]]):
        self._step = value

    @property
    def relative_step(self) -> bool:
        return self._rel_step

    @relative_step.setter
    def relative_step(self, value: bool):
        self._rel_step = value

    @property
    def minimum_step(self) -> float:
        return self._min_step

    @minimum_step.setter
    def minimum_step(self, value: float):
        self._min_step = value

    @property
    def progress_bar(self) -> bool:
        return self._pbar

    @progress_bar.setter
    def progress_bar(self, value: bool):
        self._pbar = value

    def __call__(self, wl: List[float]):
        """ See WeightingFunctionBase.__call__. """

        # get the weighting function dimensions
        box_alts = self.wf_altitudes
        box_widths = self.wf_widths

        # define the altitude grid(s) that will be used for all calculations
        box_points = np.concatenate((box_alts, box_alts + box_widths, box_alts - box_widths))
        self.append_atmosphere_grid(box_points)
        prof_alts = self.atmosphere_grid

        # allocate weighting function and radiance arrays
        self._rad = self.radiance(wl)
        if self.output_format == 'numpy':
            wf = np.zeros((len(wl), len(self.geometry.lines_of_sight), len(self.wf_species), len(self.wf_altitudes)))
        elif self.output_format == 'xarray':
            wf = self._rad
        else:
            raise ValueError('output_method must be \'numpy\' or \'xarray\'')

        pbar = None
        if self.progress_bar is True:
            nwf = len(self.wf_species) * len(self.wf_altitudes)
            pbar = tqdm(total=nwf, desc='Weighting Functions')
        elif isinstance(self.progress_bar, tqdm):
            pbar = self.progress_bar

        for i, (wf_alt, wf_width) in enumerate(zip(box_alts, box_widths)):  # loop through wf altitudes
            ialt = np.where(prof_alts == wf_alt)[0][0]

            for j in range(len(self.wf_species)):  # loop through wf species
                spec = self.wf_species[j]
                if self.output_format == 'numpy':
                    wf[:, :, j, i] = self._finite_difference(wl, self._rad, prof_alts, i, ialt, spec)
                elif self.output_format == 'xarray':
                    wf0 = self._finite_difference(wl, self._rad, prof_alts, i, ialt, spec).rename('wf_{}'.format(spec))
                    spec = 'wf_{}'.format(spec)
                    if spec in wf:
                        wf = xr.merge([wf, wf0])
                    else:
                        wf[spec] = wf0
                else:
                    raise ValueError('output_format must be \'numpy\' or \'xarray\'')
                if self.progress_bar is not False:
                    pbar.update(1)

        if self.progress_bar is True:
            pbar.close()

        self.restore_atmosphere_grid()

        # stack the altitude coord to match SASKTRAN output
        if self.output_format == 'xarray':
            wf = wf.stack(perturbation=('altitude',))

        return wf

    def _perturb(self, altitudes, profile, prof_alt_index, wf_alt_index):
        # returns the pertubation for
        step = np.atleast_1d(self.step)
        if len(step) > 1:
            step = step[wf_alt_index]
        else:
            step = step[0]

        if self.relative_step:
            step = max(self.minimum_step, step * profile[prof_alt_index])

        width = self.wf_widths[wf_alt_index]
        weights = altitudes - altitudes[prof_alt_index]
        weights = 1.0 - np.abs(weights) / width
        weights[weights < 0] = 0
        return step * weights

    def _finite_difference(self, wavelengths, true_radiance, altitudes, wf_alt_index, prof_alt_index, wf_species):
        # create perturbed atmospheres, calculate radiances and calculate weighting functions for forward,
        # backward or central finite difference method

        # fetch the reference point
        ref_pt = self.reference_point

        # store a copy of the wf_species parameter
        original_wf_species = self.atmosphere.wf_species
        self.atmosphere.wf_species = None

        # construct the perturbed atmospheres(s) and calculate perturbed radiance(s)
        wf_spec = self.atmosphere[wf_species]
        prof_true = wf_spec.climatology.get_parameter(wf_spec.species, ref_pt[0], ref_pt[1], altitudes, ref_pt[3])
        perturbation = self._perturb(altitudes, prof_true, prof_alt_index, wf_alt_index)

        # lower perturbation
        if self.method.lower() == 'forward':
            rad_perturbed_low = true_radiance
        elif self.method.lower() == 'backward' or self.method.lower() == 'central':
            prof_perturbed_low = prof_true.copy()
            prof_perturbed_low[prof_alt_index] -= perturbation
            clim_perturbed_low = sk.ClimatologyUserDefined(altitudes, {wf_spec.species: prof_perturbed_low})
            self.atmosphere[wf_species] = sk.Species(wf_spec.optical_property, clim_perturbed_low, wf_spec.species)
            rad_perturbed_low = self.radiance(wavelengths, last_wf=False)
        else:
            raise ValueError('method must be \'forward\', \'backward\' or \'central\'')

        # higher perturbation
        if self.method.lower() == 'forward' or self.method.lower() == 'central':
            prof_perturbed_high = prof_true + perturbation
            clim_perturbed_high = sk.ClimatologyUserDefined(altitudes, {wf_spec.species: prof_perturbed_high})
            self.atmosphere[wf_species] = sk.Species(wf_spec.optical_property, clim_perturbed_high, wf_spec.species)
            rad_perturbed_high = self.radiance(wavelengths, last_wf=False)
        elif self.method.lower() == 'backward':
            rad_perturbed_high = true_radiance
        else:
            raise ValueError('method must be \'forward\', \'backward\' or \'central\'')

        if self.method.lower() == 'central':
            perturbation *= 2

        # calculate the derivative
        if self.output_format == 'numpy':
            wf = (rad_perturbed_high - rad_perturbed_low) / perturbation
        elif self.output_format == 'xarray':
            wf = (rad_perturbed_high.radiance - rad_perturbed_low.radiance) / perturbation[prof_alt_index]
            wf = wf.expand_dims('altitude')
            wf.coords['altitude'] = ('altitude', [altitudes[prof_alt_index]])
        else:
            raise ValueError('output_format must be \'numpy\' or \'xarray\'')

        # restore the original atmosphere
        self.atmosphere.wf_species = original_wf_species
        self.atmosphere[wf_species] = wf_spec

        return wf
