import numpy as np
import sasktran as sk
from tqdm import tqdm

from typing import List, Union

import doas.wf.finite_difference_wf
import doas.wf.wf


class WeightingFunctionHRBase(doas.wf.wf.WeightingFunctionBase):
    """
    Base class for classes that help access and calculate weighting functions using Sasktran's High-Resolution engine.
    """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, wf_species: Union[str, List[str]], **options):
        super().__init__(geometry, atmosphere, wf_species, **options)

    @property
    def wf_altitudes(self):
        """ See WeightingFunctionFiniteDifferenceBase.wf_altitudes. """
        for key in self.options:
            if key.lower() == 'wfheights':
                return self.options[key]
        return np.linspace(500.0, 99500.0, 100)

    @property
    def wf_widths(self):
        """ See WeightingFunctionFiniteDifferenceBase.wf_widths. """
        for key in self.options:
            if key.lower() == 'wfwidths':
                return self.options[key]
        return np.ones(100) * 1000.0

    @property
    def reference_point(self):
        """ See WeightingFunctionFiniteDifferenceBase.reference_point. """
        for key in self.options.keys():
            if key.lower() is 'referencepoint':
                return self.options[key]

        engine = sk.EngineHR(self.geometry, self.atmosphere, [500.0], self.options)
        engine._add_lines_of_sight()
        engine._initialize_model()
        return engine.model_parameters['referencepoint']

    @property
    def atmosphere_grid(self) -> np.array:
        for key in self.options:
            if key.lower() == 'manualopticalheights':
                return self.options[key]
        return np.linspace(0, 1e5, 201)

    @property
    def atmosphere_grid_handle(self) -> str:
        return 'manualopticalheights'

    @property
    def surface_height(self) -> float:
        for key in self.options:
            if key.lower() == 'surfaceheight':
                return self.options[key]
        return 0.0

    @property
    def top_of_atmosphere(self) -> float:
        for key in self.options:
            if key.lower() == 'toaheight':
                return self.options[key]
        return 1e5

    def append_atmosphere_grid(self, values: np.array):
        # save the appropriate options
        grid_names = ['raytracingshells', 'solarraytracingshells', 'manualraytracingshells',
                      'manualsolarraytracingshells', 'manualopticalheights']
        pop = [name for name in self.options.keys() if name.lower() in grid_names]
        self._temp_options = {key: self.options.pop(key) for key in pop}

        # create the new manual grids
        shells = np.linspace(0, 1e5, 101)  # default
        solar = np.linspace(0, 1e5, 101)  # default
        optical = np.linspace(0, 1e5, 201)  # default
        for key, value in self._temp_options.items():
            if key.lower() == grid_names[2]:
                shells = value
            elif key.lower() == grid_names[3]:
                solar = value
            elif key.lower() == grid_names[4]:
                optical = value
        for key, value in self._temp_options.items():  # shell spacing overrides manual shells
            if key.lower() == grid_names[0]:
                shells = np.arange(self.surface_height, self.top_of_atmosphere, value)
            elif key.lower() == grid_names[1]:
                solar = np.arange(self.surface_height, self.top_of_atmosphere, value)

        # add, sort and filter the new altitudes
        shells = np.append(shells, values)
        shells[shells < self.surface_height] = self.surface_height
        shells[shells > self.top_of_atmosphere] = self.top_of_atmosphere
        shells = np.unique(shells)

        solar = np.append(solar, values)
        solar[solar < self.surface_height] = self.surface_height
        solar[solar > self.top_of_atmosphere] = self.top_of_atmosphere
        solar = np.unique(solar)

        optical = np.append(optical, values)
        optical[optical < self.surface_height] = self.surface_height
        optical[optical > self.top_of_atmosphere] = self.top_of_atmosphere
        optical = np.unique(optical)

        # set the manual grids
        self.options[grid_names[2]] = shells
        self.options[grid_names[3]] = solar
        self.options[grid_names[4]] = optical

    def restore_atmosphere_grid(self):
        if hasattr(self, '_temp_options'):
            self.options.pop('manualraytracingshells', None)
            self.options.pop('manualsolarraytracingshells', None)
            self.options.pop('manualopticalheights', None)
            for key in self._temp_options:
                self.options[key] = self._temp_options[key]
            del self._temp_options

    def radiance(self, wl: List[float], last_wf=False):
        """ See WeightingFunctionBase.radiance. """
        if last_wf is True:
            return self._rad
        else:
            temp_spec = self.atmosphere.wf_species
            self.atmosphere.wf_species = None
            self._engine = sk.EngineHR(self.geometry, self.atmosphere, wl, self.options)
            rad = self._engine.calculate_radiance(self.output_format)
            self.atmosphere.wf_species = temp_spec
            return rad


class WeightingFunctionHR(WeightingFunctionHRBase):
    """
    Class to calculate weighting functions using the analytical approximations in Sasktran's High-Resolution (HR)
    engine.
    """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, wf_species: Union[str, List[str]], **options):
        """ For parameter definitions see WeightingFunctionBase.__init__. """
        super().__init__(geometry, atmosphere, wf_species, **options)

    def __call__(self, wl: List[float]):
        """ See WeightingFunctionBase.__call__. """
        # hold a temporary copy of the atmosphere's wf_species
        temp_spec = self.atmosphere.wf_species

        # Get box altitudes (wf altitudes)
        box_alts = self.options.get('wfheights', np.linspace(500, 99500, 100))

        # Get dimensions
        nlos = len(self.geometry.lines_of_sight)
        nrs = len(self.wf_species)
        nwl = len(wl)
        nbox = len(box_alts)

        # calculate weighting functions at each los and wf species
        wf = np.zeros((nwl, nlos, nrs, nbox))
        rad = np.zeros((nwl, nlos))

        self._engine = sk.EngineHR(geometry=self.geometry, atmosphere=self.atmosphere, wavelengths=wl, options=self.options)
        if self.output_format == 'numpy':
            for j, spec in enumerate(self.wf_species):
                self.atmosphere.wf_species = spec
                result = self._engine.calculate_radiance()
                rad, wf[:, :, j, :] = result.radiance, result.weighting_function
            self._rad = rad
        elif self.output_format == 'xarray':
            for j, spec in enumerate(self.wf_species):
                self.atmosphere.wf_species = spec
                data = self._engine.calculate_radiance('xarray')
                if j == 0:
                    wf = data
                else:
                    wf['wf_{}'.format(spec)] = data['wf_{}'.format(spec)]
            self._rad = wf.radiance.values
        else:
            raise ValueError('output_format must be \'numpy\' or \'xarray\'')

        # restore the atmosphere's wf_species to its original states
        self.atmosphere.wf_species = temp_spec

        return wf


class WeightingFunctionFiniteDifferenceHR(WeightingFunctionHRBase,
                                          doas.wf.finite_difference_wf.WeightingFunctionFiniteDifferenceBase):
    """
    Class to calculate weighting functions using Sasktran's High-Resolution (HR) and finite difference approximations.
    """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, wf_species: Union[str, List[str]],
                 method: str='forward', step: Union[float, List[float]]=0.01, relative_step: bool=True,
                 minimum_step: float=1e8, progress_bar: Union[bool, tqdm]=False, **options):
        """ For parameter definitions see WeightFunctionFiniteDifferenceBase.__init__. """
        WeightingFunctionHRBase.__init__(self, geometry, atmosphere, wf_species, **options)
        doas.wf.finite_difference_wf.WeightingFunctionFiniteDifferenceBase.__init__(
            self, geometry, atmosphere, wf_species, method, step, relative_step, minimum_step, progress_bar, **options)
