import numpy as np
import sasktran as sk

from typing import List, Union

import doas.wf.finite_difference_wf
import doas.wf.wf


class WeightingFunctionMCBase(doas.wf.wf.WeightingFunctionBase):
    """
    Base class to override MC-specific properties and methods.
    """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, wf_species: Union[str, List[str]], **options):
        super().__init__(geometry, atmosphere, wf_species, **options)

    @property
    def reference_point(self):
        """ See WeightingFunctionFiniteDifferenceBase.reference_point. """
        for key in self.options.keys():
            if key.lower() is 'referencepoint':
                return self.options[key]

        engine = sk.EngineMC(self.geometry, self.atmosphere, [500.0], self.options)
        engine._add_lines_of_sight()
        engine._initialize_model()
        return engine.model_parameters['referencepoint']

    @property
    def atmosphere_grid(self) -> np.array:
        for key in self.options:
            if key.lower() == 'manualopticalheights':
                return self.options[key]
        return np.linspace(0, 1e5, 201)

    @property
    def atmosphere_grid_handle(self) -> str:
        return 'manualopticalheights'

    @property
    def surface_height(self) -> float:
        for key in self.options:
            if key.lower() == 'surfaceheight':
                return self.options[key]
        return 0.0

    @property
    def top_of_atmosphere(self) -> float:
        for key in self.options:
            if key.lower() == 'toaheight':
                return self.options[key]
        return 1e5

    def append_atmosphere_grid(self, values: np.array):
        # save the appropriate options
        grid_names = ['setnumraytracingshells', 'setsolartablealtitudedelta', 'manualraytracingshells',
                      'manualsolarraytracingshells', 'manualopticalheights']
        pop = [name for name in self.options.keys() if name.lower() in grid_names]
        self._temp_options = {key: self.options.pop(key) for key in pop}

        # get ground and top of atmosphere altitudes
        ground, toa = 0.0, 1e5
        for key, value in self.options.items():
            if key.lower() == 'surfaceheight':
                ground = value
            elif key.lower() == 'toaheight':
                toa = value

        # create the new manual grids
        shells = np.linspace(0.0, 1e5, 101)  # default
        solar = np.linspace(0.0, 1e5, 101)  # default
        optical = np.linspace(0.0, 1e5, 201)  # default
        for key, value in self._temp_options.items():
            if key.lower() == grid_names[0]:  #
                shells = np.linspace(self.surface_height, self.top_of_atmosphere, value + 1)
            elif key.lower() == grid_names[1]:
                solar = np.arange(self.surface_height, self.top_of_atmosphere, value)
        for key, value in self._temp_options.items():  # manual grids override spacing/numshells in MC
            if key.lower() == grid_names[2]:
                shells = value
            elif key.lower() == grid_names[3]:
                solar = value
            elif key.lower() == grid_names[4]:
                optical = value

        # add, sort and filter the new altitudes
        shells = np.append(shells, values)
        shells[shells < self.surface_height] = self.surface_height
        shells[shells > self.top_of_atmosphere] = self.top_of_atmosphere
        shells = np.unique(shells)

        solar = np.append(solar, values)
        solar[solar < self.surface_height] = self.surface_height
        solar[solar > self.top_of_atmosphere] = self.top_of_atmosphere
        solar = np.unique(solar)

        optical = np.append(optical, values)
        optical[optical < self.surface_height] = self.surface_height
        optical[optical > self.top_of_atmosphere] = self.top_of_atmosphere
        optical = np.unique(optical)

        # set the manual grids
        self.options[grid_names[2]] = shells
        self.options[grid_names[3]] = solar
        self.options[grid_names[4]] = optical

    def restore_atmosphere_grid(self):
        if hasattr(self, '_temp_options'):
            self.options.pop('manualraytracingshells', None)
            self.options.pop('manualsolarraytracingshells', None)
            self.options.pop('manualopticalheights', None)
            for key in self._temp_options:
                self.options[key] = self._temp_options[key]
            del self._temp_options


class WeightingFunctionMCNull(WeightingFunctionMCBase):
    """
    This class is just to keep the AMF class structure consistent (HR and DO use WFs to calculated AMFs, but MC does
    not). It does not actually calculate any weighting functions, it just calculates radiances and sets/retrieves
    some parameters.
    """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, **options):
        super().__init__(geometry, atmosphere, [], **options)

    def __call__(self, wl: List[float]):
        raise NotImplementedError()

    def radiance(self, wl: List[float], last_wf=False):
        """ See WeightingFunctionBase.radiance. """
        if last_wf is True:
            return self._rad
        else:
            self._engine = sk.Engine('MC', self.geometry, self.atmosphere, wl, self.options)
            self._rad = self._engine.calculate_radiance(self.output_format)
            return self._rad
