import sasktran as sk
import numpy as np

from typing import List, Union
from scipy.interpolate import interp1d


class WeightingFunctionBase(object):
    """
    Base class for classes that help access and calculate weighting functions using different methods and different
    engines from Sasktran.
    """
    def __init__(self, geometry: sk.Geometry, atmosphere: sk.Atmosphere, wf_species: Union[str, List[str]], **options):
        """
        Parameters
        ----------
        geometry: sk.Geometry
            Geometry object from sasktran containing information about the lines of sight and the sun.
        atmosphere: sk.Atmosphere
            Atmosphere object from sasktran containing information about the atmosphere. If atmosphere.wf_species
            is set, it will be overruled by wf_species.
        wf_species: str, List[str]
            List of species corresponding to the names of species in atmosphere for which weighting functions will be
            calculated.
        **options
            Options that are passed to the engine.
        """
        self.atmosphere = atmosphere
        self.geometry = geometry
        self.wf_species = wf_species
        self.options = options
        self.output_format = 'xarray'
        self._rad = None
        self._engine = None

    @property
    def wf_species(self) -> List[str]:
        return self._wf_species

    @wf_species.setter
    def wf_species(self, species: Union[str, List[str]]):
        self._wf_species = list(np.atleast_1d(species))

    @property
    def atmosphere(self):
        return self._atmosphere

    @atmosphere.setter
    def atmosphere(self, atmosphere: sk.Atmosphere()):
        self._atmosphere = atmosphere

    @property
    def geometry(self):
        return self._geometry

    @geometry.setter
    def geometry(self, geometry: sk.Geometry()):
        self._geometry = geometry

    @property
    def options(self):
        return self._options

    @options.setter
    def options(self, options: dict):
        self._options = options

    @property
    def engine(self):
        return self._engine

    @property
    def output_format(self):
        return self._out

    @output_format.setter
    def output_format(self, value):
        if value.lower() == 'numpy':
            self._out = 'numpy'
        elif value.lower() == 'xarray':
            self._out = 'xarray'
        else:
            raise ValueError('output_format must be \'numpy\' or \'xarray\'')

    @property
    def wf_altitudes(self) -> np.array:
        """ Fetches the weighting function altitudes from the engine options, or returns the default if it is
        unchanged. """
        raise NotImplementedError()

    @property
    def wf_widths(self) -> np.array:
        """ Fetches the weighting function widths from the engine options, or returns the default if it is
        unchanged. """
        raise NotImplementedError()

    @property
    def reference_point(self) -> np.array:
        """ Fetches the reference point in form np.array([latitude, longitude, altitude, mjd]). """
        raise NotImplementedError()

    @property
    def atmosphere_grid(self) -> np.array:
        """ Fetches the points where profiles are defined in metres. """
        raise NotImplementedError()

    @property
    def top_of_atmosphere(self) -> float:
        """ Fetches the top of atmosphere in metres. """
        raise NotImplementedError()

    @property
    def surface_height(self) -> float:
        """ Fetches the surface height in metres. """
        raise NotImplementedError()

    def append_atmosphere_grid(self, values: np.array):
        """ Adds values to the atmospheric grid set in self.options (if it exists). Should always be undone by a
        subsequent call to restore_atmosphere_grid. """
        raise NotImplementedError()

    def restore_atmosphere_grid(self):
        """ Restores the atmospheric grid to its original value after it has been altered by append_atmospher_grid. """
        raise NotImplementedError()

    def radiance(self, wl: List[float], last_wf=False):
        """
        Calculate or fetch radiance values. Weighting function calculation should be turned off when the radiance
        is calculated here.

        Parameters
        ----------
        wl: List[float]
            Wavelengths in nm to calculate the radiances at.
        last_wf: bool, False
            If True, the radiance from the last wf calculation is retrieved (this may not correspond to the given wl).
            If False, the radiance is calculated.

        Returns
        -------
        radiance: np.ndarray, xr.Dataset
            Calculated or fetched radiances. Has shape (len(wavelengths), len(lines of sight)), where wavelengths refers
            to the wavelengths used in the last wf call if last_wf is True.
        """
        raise NotImplementedError()

    def __call__(self, wl: List[float]):
        """
        Calculates the weighting functions.

        Parameters
        ----------
        wl: List[float]
            Wavelengths in nm to calculate weighting functions at.

        Returns
        -------
        wf: np.ndarray
            Weighting functions calculated for each wavelength, line of sight, wf species and altitude. Has shape
            (len(wavelengths), len(lines of sight), len(wf species), len(wf altitudes)).
        """
        raise NotImplementedError()


class WeightingFunctionCustom(WeightingFunctionBase):
    """ Used to store and retrieve weighting functions without calculation. """
    def __init__(self, wavelengths: List[float], weighting_functions: np.ndarray,
                 interpolation_method: Union[str, int]='linear', geometry: sk.Geometry=None,
                 atmosphere: sk.Atmosphere=None, wf_species=None, **options):
        super().__init__(geometry, atmosphere, wf_species=wf_species, **options)
        self._wl = wavelengths
        self._wf = weighting_functions
        self._interp_method = interpolation_method
        self._create_interp()

    def _create_interp(self):
        self._interp = interp1d(self._wl, self._wf, kind=self._interp_method, axis=0)

    @property
    def wavelengths(self):
        return self._wl

    @wavelengths.setter
    def wavelengths(self, value: List[float]):
        self._wl = value
        self._create_interp()

    @property
    def weighting_functions(self):
        return self._wf

    @weighting_functions.setter
    def weighting_functions(self, value: np.ndarray):
        self._wf = value
        self._create_interp()

    @property
    def interpolation_method(self):
        return self._interp_method

    @interpolation_method.setter
    def interpolation_method(self, value: Union[str, int]):
        self._interp_method = value
        self._create_interp()

    def __call__(self, wl: List[float]) -> np.ndarray:
        return self._interp(wl)
